/*
	Copyright  1995-2001, The AROS Development Team. All rights reserved.
	Copyright  2006-2007, Harry Sintonen. All rights reserved.
	$Id: eventfuncs.c,v 1.4 2012/02/15 01:00:54 jacadcaps Exp $
 
	Desc: ...
	Lang: English.
*/


#ifndef COMMODITIES_BASE_H
#include "cxintern.h"
#endif
#ifndef PROTO_EXEC_H
#include <proto/exec.h>
#endif
#ifndef PROTO_UTILITY_H
#include <proto/utility.h>
#endif
#ifndef EXEC_MEMORY_H
#include <exec/memory.h>
#endif
#ifndef DEVICES_INPUTEVENT_H
#include <devices/inputevent.h>
#endif
#ifdef __MORPHOS__
#include <hardware/atomic.h>
#endif

void memclr(void *, int);


/*
 * Ultrafast object cache code (TM)  2006 Harry Sintonen
 *
 * At the moment the objects are never deallocated, except at the library
 * expunge. Typically the memory usage settles around 4k-6k of memory
 * used.
 */

#define DEBUG_MEMUSAGE		0


struct SNode
{
	struct SNode *succ;
};

struct SList
{
	struct SNode *head;
};

struct objcache
{
	struct SList freelist;
	struct SList puddlelist;
	ULONG        objsize;
};

#define NUMPUDDLEOBJ 8  /* number of objects to allocate per puddle */

struct puddle
{
	struct SNode node;
	/* NUMPUDDLEOBJ * objsize bytes of data follows */
};

#ifdef __MORPHOS__
/* Oi, perfect use for atomic singly linked lists. w00t \o/ - Piru */
#define S_LOCKLIST(list) do {} while (0)
#define S_UNLOCKLIST(list) do {} while (0)
#define S_NEWLIST(list) ATOMIC_STORE((LONG *) &(list)->head, 0)
#define S_ADDHEAD(list,node) ATOMIC_SADDHEAD((APTR *) (list), node)
#define S_REMHEAD(list) ATOMIC_SREMHEAD((APTR *) (list))
#else
#define S_LOCKLIST(list) Forbid()
#define S_UNLOCKLIST(list) Permit()
#define S_NEWLIST(list) do { (list)->head = NULL; } while (0)
#define S_ADDHEAD(list,node) do { struct SList *__l = (list); struct SNode *__n = (node); __n->succ = __l->head; __l->head = __n; } while (0)
#define S_REMHEAD(list) ({ struct SList *__l = (list); struct SNode *__n; __n = __l->head; if (__n) { __l->head = __n->succ; } (APTR)__n; })
#endif

static struct objcaches
{
	struct objcache c_cx_common;
	struct objcache c_cx_filter;
	struct objcache c_cx_send;
	struct objcache c_cx_signal;
	struct objcache c_cx_broker;
	struct objcache c_cx_custom;
	struct objcache c_cx_message;
	struct objcache c_cx_inputevent;
	struct objcache c_iecnpp_pixel;
	struct objcache c_iecnpp_tablet;
	struct objcache c_iecnpp_newtablet;
} objcaches =
{
#define OCI(s) { {NULL}, {NULL}, ((s) + 3) & ~3 }

	OCI(sizeof(CxObj)),
	OCI(sizeof(CxObj) + sizeof(struct InputXpression)),
	OCI(sizeof(CxObj) + sizeof(struct SendExt)),
	OCI(sizeof(CxObj) + sizeof(struct SignalExt)),
	OCI(sizeof(CxObj) + sizeof(struct BrokerExt)),
	OCI(sizeof(CxObj) + sizeof(struct CustomExt)),
	OCI(sizeof(CxMsg)),
	OCI(sizeof(struct GeneratedInputEvent)),
	OCI(sizeof(struct IEPointerPixel)),
	OCI(sizeof(struct IEPointerTablet)),
	OCI(sizeof(struct IENewTablet))
};
static const int objcachescnt = sizeof(struct objcaches) / sizeof(struct objcache);


static struct objcache * const objcache[] =
{
	&objcaches.c_cx_common,	/* CX_INVALID */
	&objcaches.c_cx_filter,	/* CX_FILTER */
	&objcaches.c_cx_common,	/* CX_TYPEFILTER */
	&objcaches.c_cx_send,	/* CX_SEND */
	&objcaches.c_cx_signal,	/* CX_SIGNAL */
	&objcaches.c_cx_common,	/* CX_TRANSLATE */
	&objcaches.c_cx_broker,	/* CX_BROKER */
	&objcaches.c_cx_common,	/* CX_DEBUG */
	&objcaches.c_cx_custom,	/* CX_CUSTOM */
	&objcaches.c_cx_common 	/* CX_ZERO */
};


VOID oc_init(struct Library *CxBase)
{
}


/* oc_cleanup is called within Forbid() (at expunge)...
*/
VOID oc_cleanup(struct Library *CxBase)
{
	int i;

	/* This is a bit ugly, but oh well... */
	for (i = 0; i < objcachescnt; i++)
	{
		struct objcache *cache = &((struct objcache *)&objcaches)[i];
		const int objsize = cache->objsize;
		struct puddle *puddle;

		while ((puddle = S_REMHEAD(&cache->puddlelist)))
		{
			FreeMem(puddle, sizeof(*puddle) + NUMPUDDLEOBJ * objsize);
		}
		S_NEWLIST(&cache->freelist);
	}
}


static APTR oc_alloc(struct objcache *cache, struct Library *CxBase)
{
	APTR obj;

	S_LOCKLIST(&cache->freelist);
	obj = S_REMHEAD(&cache->freelist);
	S_UNLOCKLIST(&cache->freelist);
	if (!obj)
	{
		const int objsize = cache->objsize;
		struct puddle *puddle;

		puddle = AllocMem(sizeof(*puddle) + NUMPUDDLEOBJ * objsize, MEMF_ANY);
		if (puddle)
		{
#if NUMPUDDLEOBJ > 1
			int i;

			obj = puddle + 1;
			S_LOCKLIST(&cache->puddlelist);
			S_ADDHEAD(&cache->puddlelist, (APTR) puddle);
			for (i = 0; i < (NUMPUDDLEOBJ - 1); i++)
			{
				S_ADDHEAD(&cache->freelist, (APTR) obj);

				obj = ((char *) obj) + objsize;
			}
#else
			obj = puddle + 1;
			S_LOCKLIST(&cache->puddlelist);
			S_ADDHEAD(&cache->puddlelist, puddle);
#endif

#if DEBUG_MEMUSAGE
			{
				static ULONG total = 0;
				total += (sizeof(*puddle) + NUMPUDDLEOBJ * objsize + MEM_BLOCKMASK) & ~MEM_BLOCKMASK;
				dprintf("cx: total %lu bytes object cache memory used\n", total);
			}
#endif
			S_UNLOCKLIST(&cache->puddlelist);
		}
	}

	return obj;
}


static APTR oc_allocclr(struct objcache *cache, struct Library *CxBase)
{
	APTR obj;

	obj = oc_alloc(cache, CxBase);
	if (obj)
	{
		memclr(obj, cache->objsize);
	}

	return obj;
}


static VOID oc_free(struct objcache *cache, APTR obj, struct Library *CxBase)
{
	if (obj)
	{
		S_LOCKLIST(&cache->freelist);
		S_ADDHEAD(&cache->freelist, (APTR) obj);
		S_UNLOCKLIST(&cache->freelist);
	}
}


#if 0
static APTR ob_alloc(int objtype, struct Library *CxBase)
{
	if ((unsigned int) objtype > CX_ZERO)
	{
		return NULL;
	}

	return oc_alloc(objcache[objtype], CxBase);
}
#endif


static APTR ob_allocclr(int objtype, struct Library *CxBase)
{
	if ((unsigned int) objtype > CX_ZERO)
	{
		return NULL;
	}

	return oc_allocclr(objcache[objtype], CxBase);
}


static VOID ob_free(int objtype, APTR obj, struct Library *CxBase)
{
	if ((unsigned int) objtype > CX_ZERO)
	{
		return;
	}

	oc_free(objcache[objtype], obj, CxBase);
}


APTR AllocCxStructure(LONG type, LONG objtype, struct Library *CxBase)
{
	APTR   temp = NULL;
	CxObj *tempObj;
	CxMsg *tempMsg;

	switch (type)
	{
	case CX_OBJECT:

		tempObj = ob_allocclr(objtype, CxBase);

		if (tempObj)
		{
			CXOBJType(tempObj) = objtype;
			tempObj->co_Ext.co_FilterIX = (APTR)(tempObj + 1);
			NEWLIST(&tempObj->co_ObjList);

			/* This is done to make it easy for Exchange */
			if (objtype == CX_BROKER)
			{
				tempObj->co_Node.ln_Name = (char *)&tempObj->co_Ext.co_BExt->bext_Name;
			}
			temp = (APTR)tempObj;
		}

		break;

	case CX_MESSAGE:

		switch (objtype)
		{
		case CXM_SINGLE:

			tempMsg = oc_allocclr(&objcaches.c_cx_message, CxBase);

			if (tempMsg)
			{
				tempMsg->cxm_Type = CXM_COMMAND;
				//tempMsg->cxm_Data = NULL;
				tempMsg->cxm_Message.mn_ReplyPort = &GPB(CxBase)->cx_MsgPort;
				tempMsg->cxm_Message.mn_Length = sizeof(CxMsg);
				temp = (APTR)tempMsg;
			}

			break;

		case CXM_DOUBLE:

			tempMsg = oc_allocclr(&objcaches.c_cx_message, CxBase);

			if (tempMsg)
			{
				if ((tempMsg->cxm_Data = AllocCxStructure(CX_INPUTEVENT, 0, CxBase)))
				{
					tempMsg->cxm_Type = CXM_IEVENT;
					tempMsg->cxm_Message.mn_ReplyPort = &GPB(CxBase)->cx_MsgPort;
					tempMsg->cxm_Message.mn_Length = sizeof(CxMsg);

					temp = (APTR)tempMsg;
				}
				else
				{
					oc_free(&objcaches.c_cx_message, tempMsg, CxBase);
				}
			}

			break;
		}

		break;

	case CX_INPUTEVENT:

		/* No need to oc_allocclr as long as always
		 * CopyInputEvent() to. This is the case atm.
		 */
		temp = oc_alloc(&objcaches.c_cx_inputevent, CxBase);

		break;
	}

	return temp;
}


VOID FreeCxStructure(APTR obj, int type, struct Library *CxBase)
{
	if (obj == NULL)
	{
		return;
	}

	switch (type)
	{
	case CX_OBJECT:
		ob_free(CXOBJType((CxObj *)obj), obj, CxBase);
		break;

	case CX_MESSAGE:
		FreeCxStructure(((CxMsg *)obj)->cxm_Data, CX_INPUTEVENT, CxBase);
		oc_free(&objcaches.c_cx_message, obj, CxBase);
		break;

	case CX_INPUTEVENT:
		{
			struct InputEvent *ie = (APTR) obj;
			struct GeneratedInputEvent *gie = (APTR)obj;

			/* someone might have zapped the ie->ie_EventAddress so we rely on a extra copy */
			if (gie->Class == IECLASS_NEWPOINTERPOS &&
			    gie->EventAddress)
			{
				switch (gie->SubClass)
				{
				case IESUBCLASS_PIXEL:
					oc_free(&objcaches.c_iecnpp_pixel, gie->EventAddress, CxBase);
					break;

				case IESUBCLASS_TABLET:
					oc_free(&objcaches.c_iecnpp_tablet, gie->EventAddress, CxBase);
					break;

				case IESUBCLASS_NEWTABLET:
					{
						struct IENewTablet *ient;

						if ((ient = (APTR)gie->EventAddress))
						{
							if (ient->ient_TagList)
							{
								FreeTagItems(ient->ient_TagList);
							}
							oc_free(&objcaches.c_iecnpp_newtablet, ient, CxBase);
						}
					}
					break;
				}

				gie->EventAddress = NULL;
				gie->Class = -1;
			}
		}
		oc_free(&objcaches.c_cx_inputevent, obj, CxBase);
		break;
	}
}


BOOL CopyInputEvent(struct InputEvent *from, struct InputEvent *to,
                    struct Library *CxBase)
{
	struct GeneratedInputEvent *gie = (APTR)to;

	CopyMem(from, to, sizeof(struct InputEvent));

	gie->Class = from->ie_Class;
	gie->SubClass = from->ie_SubClass;

	if (from->ie_Class == IECLASS_NEWPOINTERPOS &&
	    from->ie_EventAddress)
	{
		switch (from->ie_SubClass)
		{
		case IESUBCLASS_PIXEL:
			if ((to->ie_EventAddress = oc_alloc(&objcaches.c_iecnpp_pixel, CxBase)) == NULL)
			{
				gie->EventAddress = to->ie_EventAddress;
				return FALSE;
			}

			CopyMem(from->ie_EventAddress, to->ie_EventAddress, sizeof(struct IEPointerPixel));
			gie->EventAddress = to->ie_EventAddress;
			break;

		case IESUBCLASS_TABLET:
			if ((to->ie_EventAddress = oc_alloc(&objcaches.c_iecnpp_tablet, CxBase)) == NULL)
			{
				gie->EventAddress = to->ie_EventAddress;
				return FALSE;
			}

			CopyMem(from->ie_EventAddress, to->ie_EventAddress, sizeof(struct IEPointerTablet));
			gie->EventAddress = to->ie_EventAddress;
			break;

		case IESUBCLASS_NEWTABLET:
			{
				struct IENewTablet *fromient, *toient;
				struct TagItem *taglist;

				if ((to->ie_EventAddress = toient = oc_alloc(&objcaches.c_iecnpp_newtablet, CxBase)) == NULL)
				{
					gie->EventAddress = to->ie_EventAddress;
					return FALSE;
				}

				fromient = (APTR)from->ie_EventAddress;
				CopyMem(fromient, toient, sizeof(struct IENewTablet));
				gie->EventAddress = to->ie_EventAddress;

				/* Only call CloneTagItems() if really have to */
				taglist = fromient->ient_TagList;
				if (taglist && taglist->ti_Tag != TAG_DONE)
				{
					/* Just use empty taglist if this fails */
					toient->ient_TagList = CloneTagItems(taglist);
				}
				else
				{
					toient->ient_TagList = NULL;
				}
			}
			break;
		}
	}

	return TRUE;
}

